#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import warnings
from types import TracebackType
from typing import Type, TypeVar
from urllib.parse import urlparse

from pydantic import ValidationError

from evo import logging
from evo.common import APIConnector, RequestMethod
from evo.common.exceptions import EvoAPIException
from evo.common.interfaces import ITransport

from .data import AccessToken, DeviceFlowResponse, OAuthScopes, OIDCConfig
from .exceptions import OAuthError, OIDCError

logger = logging.getLogger("oauth")

__all__ = ["OIDCConnector"]

T = TypeVar("T", bound=AccessToken)


class OIDCConnector:
    """OpenID Connect client with endpoint discovery.

    This client is used to authenticate with an OpenID Connect provider and obtain an access token.
    OAuth 2.0 endpoints are discovered using the OpenID Connect Discovery protocol.

    https://openid.net/specs/openid-connect-discovery-1_0.html
    """

    def __init__(
        self,
        transport: ITransport,
        oidc_issuer: str,
        client_id: str,
        client_secret: str | None = None,
    ) -> None:
        """
        :param transport: The transport to use for making requests.
        :param oidc_issuer: The OpenID Connect issuer URL.
        :param redirect_url: The local URL to redirect the user back to after authorisation.
        :param client_id: The OAuth client ID, as registered with the OAuth provider.
        :param client_secret: The OAuth client secret, as registered with the OAuth provider.

        :raises OIDCError: If the issuer URL is invalid.
        """
        if not isinstance(oidc_issuer, str):
            raise OIDCError("OIDC issuer must be a string.")

        issuer = urlparse(oidc_issuer)
        if issuer.scheme != "https":
            raise OIDCError("OIDC issuer must use HTTPS.")

        if issuer.port is not None:
            raise OIDCError("OIDC issuer must not specify a port.")

        if issuer.query != "":
            raise OIDCError("OIDC issuer must not specify a query.")

        if issuer.fragment != "":
            raise OIDCError("OIDC issuer must not specify a fragment.")

        if issuer.path.endswith("/.well-known/openid-configuration"):
            issuer = issuer._replace(path=issuer.path.removesuffix("/.well-known/openid-configuration"))
            warnings.warn(
                f"OIDC issuer should not include the .well-known path. Assuming the issuer is {issuer.geturl()}."
            )

        self._connector = APIConnector(issuer.geturl(), transport)  # No authorization for OIDC discovery.
        self.__client_id = client_id
        self.__client_secret = client_secret

        self._config: OIDCConfig | None = None

    @property
    def config(self) -> OIDCConfig:
        return self._config

    @property
    def issuer(self) -> str:
        return self._connector.base_url.rstrip("/")

    @property
    def client_id(self) -> str:
        return self.__client_id

    async def __aenter__(self) -> OIDCConnector:
        await self._connector.open()
        await self.load_config()
        return self

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_value: Exception | None, traceback: TracebackType | None
    ) -> None:
        await self._connector.close()

    async def load_config(self):
        """Fetch the OIDC configuration from the issuer.

        https://openid.net/specs/openid-connect-discovery-1_0.html

        :raises OIDCError: If the configuration cannot be fetched or is invalid.
        """
        if self.config is not None:
            logger.debug("OpenID Connect configuration has already been fetched.")
            return

        logger.info(f"Fetching OpenID Connect configuration from {self.issuer}...")
        async with self._connector:
            try:
                self._config = await self._connector.call_api(
                    RequestMethod.GET,
                    ".well-known/openid-configuration",
                    header_params={"Accept": "application/json"},
                    response_types_map={"200": OIDCConfig},
                )

                if self.config.issuer != self.issuer:
                    raise OIDCError("OIDC issuer does not match the issuer in the configuration.")
            except EvoAPIException as e:
                raise OIDCError("Failed to fetch OIDC configuration.") from e

    async def begin_device_flow(self, scopes: OAuthScopes) -> DeviceFlowResponse:
        if self.config.device_authorization_endpoint is None:
            raise OIDCError("Device flow authorization is not supported by this auth provider.")

        data = {
            "client_id": self.client_id,
            "scope": str(scopes),
        }
        if self.__client_secret is not None:
            data["client_secret"] = self.__client_secret
        try:
            async with self._connector:
                try:
                    response = await self._connector.call_api(
                        RequestMethod.POST,
                        self.config.device_authorization_endpoint,
                        header_params={
                            "Accept": "application/json",
                            "Content-Type": "application/x-www-form-urlencoded",
                        },
                        post_params=data,
                        response_types_map={"200": DeviceFlowResponse},
                    )
                except EvoAPIException as e:
                    error_json: dict = e.content
                    title = error_json.get("error", "Unexpected response from server")
                    detail = error_json.get("error_description", str(e))
                    raise OAuthError(f"{title}: {detail}")
                except ValidationError as e:
                    raise OAuthError("Invalid device flow response from server.") from e
            return response
        except OAuthError:
            raise  # Re-raise OAuthError exceptions.
        except Exception as exc:  # noqa: E722
            # Catch any other exceptions and raise OAuthError.
            raise OAuthError("Unable to begin device flow.") from exc

    async def fetch_token(self, data: dict, expected_response_model: Type[T]) -> T:
        """Fetch an access token from the server.

        :param data: The data to send to the server.

        :return: The token response.

        :raises OAuthError: If the token response cannot be fetched.
        """
        data["client_id"] = self.client_id
        if self.__client_secret is not None:
            data["client_secret"] = self.__client_secret

        try:
            async with self._connector:
                try:
                    response = await self._connector.call_api(
                        RequestMethod.POST,
                        self.config.token_endpoint,
                        header_params={
                            "Accept": "application/json",
                            "Content-Type": "application/x-www-form-urlencoded",
                        },
                        post_params=data,
                        response_types_map={"200": expected_response_model},
                    )
                except EvoAPIException as e:
                    error_json: dict = e.content
                    title = error_json.get("error", "Unexpected response from server")
                    detail = error_json.get("error_description", str(e))
                    raise OAuthError(f"{title}: {detail}")
                except ValidationError as e:
                    raise OAuthError("Invalid token response from server.") from e
            return response
        except OAuthError:
            raise  # Re-raise OAuthError exceptions.
        except Exception as exc:  # noqa: E722
            # Catch any other exceptions and raise OAuthError.
            raise OAuthError("Unable to fetch access token." + str(exc)) from exc



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import json
from asyncio.log import logger
from base64 import urlsafe_b64decode
from datetime import datetime, timedelta, timezone
from enum import Flag, auto
from typing import Literal, Optional
from urllib.parse import urlparse

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from evo.common.utils import BackoffLinear, Retry
from evo.oauth.exceptions import OAuthError, OIDCError

__all__ = [
    "AccessToken",
    "DeviceFlowResponse",
    "OAuthScopes",
    "UserAccessToken",
]

_ALLOWABLE_CLOCK_DRIFT = timedelta(minutes=5)


def _utcnow() -> datetime:
    """Get the current UTC time.

    :return: The current UTC time.
    """
    return datetime.now(timezone.utc)


class OAuthScopes(Flag):
    """Public OAuth scopes for authenticating against Evo APIs.

    https://developer.seequent.com/docs/guides/getting-started/apps-and-tokens#about-evo-access-tokens
    """

    # OpenID Scopes
    openid = auto()
    profile = auto()
    organization = auto()
    email = auto()
    address = auto()
    phone = auto()

    offline_access = auto()

    # Evo Scopes
    evo_discovery = auto()
    """Required for access to the Evo Discovery API, both discovery and token endpoints."""

    evo_workspace = auto()
    """Required for access to the Workspaces API."""

    evo_blocksync = auto()
    """Required for access to the BlockSync API."""

    evo_object = auto()
    """Required for access to the Geoscience Object API."""

    evo_file = auto()
    """Required for access to the File API."""

    # Useful combinations
    """Default scopes for Evo applications.

    Extra scopes to be added as needed include `evo_blocksync`, `evo_geoscienceobject`, or `evo_file`.
    """
    default = openid | profile | organization | email | evo_discovery | evo_workspace

    """All scopes for Evo applications."""
    all_evo = default | evo_blocksync | evo_object | evo_file

    def __str__(self) -> str:
        """Format a space-separated list of scopes for the OAuth request."""
        scopes = sorted([scope for scope in self.__class__ if scope in self and len(scope) == 1], key=lambda s: s.value)
        return " ".join([scopes_map[scope] for scope in scopes])

    def __len__(self) -> int:
        """Count the number of unique scopes that are set."""
        return len([scope for scope in self.__class__ if scope in self and (scope.value & (scope.value - 1)) == 0])


scopes_map = {
    OAuthScopes.openid: "openid",
    OAuthScopes.profile: "profile",
    OAuthScopes.organization: "organization",
    OAuthScopes.email: "email",
    OAuthScopes.address: "address",
    OAuthScopes.phone: "phone",
    OAuthScopes.offline_access: "offline_access",
    OAuthScopes.evo_discovery: "evo.discovery",
    OAuthScopes.evo_workspace: "evo.workspace",
    OAuthScopes.evo_blocksync: "evo.blocksync",
    OAuthScopes.evo_object: "evo.object",
    OAuthScopes.evo_file: "evo.file",
}


class AccessToken(BaseModel):
    """A base model for token responses for an OAuth server.
    https://www.rfc-editor.org/rfc/rfc6749#section-5.1
    """

    token_type: Literal["Bearer"]
    access_token: str
    """The access token issued by the authorization server."""

    expires_in: Optional[int] = None
    """The lifetime in seconds of the access token."""

    issued_at: datetime = Field(default_factory=_utcnow)
    """The time at which the token response was received."""

    scope: Optional[str] = None
    """The scope of the access token."""

    @property
    def expires_at(self) -> datetime | None:
        """The time at which the token expires, or None if the token lifetime is unknown.

        A small margin of error may occur due to the time taken to process the token response.
        """
        if self.expires_in is None:
            return None
        return self.issued_at + timedelta(seconds=self.expires_in)

    @property
    def ttl(self) -> int | None:
        """The time-to-live (TTL) of this access token in seconds, or None if the token lifetime is unknown.

        If the token is expired, the TTL will be 0.
        """
        if self.expires_at is None:
            return None
        ttl = self.expires_at - _utcnow()
        return max(round(ttl.total_seconds()), 0)

    @property
    def is_expired(self) -> bool:
        """Whether this access token has expired.

        True if the access token has expired, False otherwise.
        If the token lifetime is unknown, this method will always return False.
        """
        if (expiry := self.expires_at) is not None:
            return _utcnow() > expiry
        else:
            return False


class UserAccessToken(AccessToken):
    """A bearer token response from an OAuth server, for authorication_grant flows.

    https://www.rfc-editor.org/rfc/rfc6749#section-5.1
    """

    id_token: str
    """The ID token issued by the authorization server."""

    refresh_token: Optional[str] = None
    """The refresh token, which can be used to obtain new access tokens using the same authorization grant."""

    def validate_id_token(self, issuer: str, client_id: str, ignore_iat: bool = False) -> None:
        """Validate an ID token according to the OpenID Connect specification.

        https://openid.net/specs/openid-connect-basic-1_0.html#IDTokenValidation

        :param issuer: The issuer URL to validate against.
        :param client_id: The client ID to validate against.
        :param ignore_iat: Whether to ignore the `iat` claim when validating the token. Sometimes when caching tokens
            it may be preferable to validate the token without checking the `iat` claim, otherwise the token may be
            rejected if it was issued more than 5 minutes ago.

        :raises OAuthError: If the ID token is invalid or cannot be decoded.
        """
        # The Client MUST validate the ID Token in the Token Response. To do this, the Client can split the ID Token at
        # the period (".") characters, take the second segment, and base64url decode it to obtain a JSON object containing
        # the ID Token Claims, which MUST be validated as follows:
        try:  # Broad exception handling to catch any errors during id_token validation.
            # Add surplus padding to make sure this can be decoded.
            # https://stackoverflow.com/a/49459036
            id_token = self.id_token.split(".")[1] + "=="
            claims: dict = json.loads(urlsafe_b64decode(id_token).decode("utf-8"))

            # 1. The Issuer Identifier for the OpenID Provider (which is typically obtained during Discovery) MUST exactly
            #    match the value of the iss (issuer) Claim.
            parsed_issuer = urlparse(issuer)
            if parsed_issuer.hostname and parsed_issuer.hostname.endswith(".seequent.com"):
                # Seequent ID is not fully compliant with OpenID Connect.
                logger.debug("Skipping issuer validation for Seequent ID.")
            elif claims.get("iss") != issuer:
                raise OAuthError("Issuer identifier does not match the value of the iss Claim.")

            # 2. The Client MUST validate that the aud (audience) Claim contains its client_id value registered at the Issuer
            #    identified by the iss (issuer) Claim as an audience. The ID Token MUST be rejected if the ID Token does not
            #    list the Client as a valid audience, or if it contains additional audiences not trusted by the Client.
            if client_id not in claims.get("aud"):  # `aud` may be a list of string, or a single string.
                raise OAuthError("Audience claim does not contain the client_id value.")

            # 3. The current time MUST be before the time represented by the exp Claim (possibly allowing for some small
            #    leeway to account for clock skew).
            exp_claim = datetime.fromtimestamp(claims.get("exp"), tz=timezone.utc) + _ALLOWABLE_CLOCK_DRIFT
            if self.issued_at > exp_claim and not ignore_iat:
                raise OAuthError("Token has expired.")

            # 4. The iat Claim can be used to reject tokens that were issued too far away from the current time, limiting the
            #    amount of time that nonces need to be stored to prevent attacks. The acceptable range is Client specific.
            min_iat = self.issued_at - _ALLOWABLE_CLOCK_DRIFT
            max_iat = self.issued_at + _ALLOWABLE_CLOCK_DRIFT
            iat_claim = datetime.fromtimestamp(claims.get("iat"), tz=timezone.utc)
            if not min_iat < iat_claim < max_iat and not ignore_iat:
                raise OAuthError("Token was issued too far away from the current time.")
        except OAuthError:
            raise  # Re-raise OAuthError exceptions.
        except Exception as e:
            raise OAuthError("Unable to decode ID token.") from e


class OIDCConfig(BaseModel):
    issuer: str
    """The issuer URL."""

    authorization_endpoint: str
    """The authorization endpoint, relative to the issuer URL."""

    token_endpoint: str
    """The token endpoint, relative to the issuer URL."""

    device_authorization_endpoint: str | None = None
    """The device authorization endpoint, relative to the issuer URL."""

    end_session_endpoint: str | None = None
    """The end session endpoint, relative to the issuer URL."""

    response_types_supported: set[str]
    """The supported response types for the authorization endpoint."""

    grant_types_supported: set[str] = set(["authorization_code", "implicit"])
    """The supported grant types for the token endpoint."""

    @field_validator(
        "authorization_endpoint",
        "token_endpoint",
        "device_authorization_endpoint",
        "end_session_endpoint",
        check_fields=True,
    )
    @classmethod
    def endpoints_exist_under_issuer(cls, v, info: ValidationInfo):
        if "issuer" in info.data:
            if not v.startswith(info.data["issuer"]):
                raise OIDCError(f"OIDC field {info.field_name} must be a URL under the issuer")
            validated_endpoint = v.removeprefix(info.data["issuer"])
            logger.info(f"Found OAuth {info.field_name} endpoint: {validated_endpoint}")
            return validated_endpoint
        else:
            raise OIDCError("Missing issuer url in OIDC config.")


class DeviceFlowResponse(BaseModel):
    """A response from the OAuth server to a device flow request.

    https://datatracker.ietf.org/doc/html/rfc8628#section-3.2
    """

    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: Optional[str] = None
    expires_in: int
    interval: int = 5

    def __str__(self) -> str:
        return f"Verification URL: {self.verification_uri}\nUser code: {self.user_code}"

    @property
    def _retry(self) -> Retry:
        max_attempts = self.expires_in // self.interval
        return Retry(logger=logger, max_attempts=max_attempts, backoff_method=BackoffLinear(self.interval))



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import contextlib
from collections.abc import AsyncIterator
from typing import Generic, TypeVar

from evo import logging
from evo.common.data import HTTPHeaderDict
from evo.common.exceptions import RetryError
from evo.common.interfaces import IAuthorizer

from .data import AccessToken, DeviceFlowResponse, OAuthScopes, UserAccessToken
from .exceptions import OAuthError, OIDCError
from .oauth_redirect_handler import OAuthRedirectHandler
from .oidc import OIDCConnector

__all__ = [
    "AuthorizationCodeAuthorizer",
    "ClientCredentialsAuthorizer",
    "DeviceFlowAuthorizer",
]

logger = logging.getLogger("oauth")

T = TypeVar("T", bound=AccessToken)


class _BaseAuthorizer(IAuthorizer, Generic[T]):
    pi_partial_implementation = True  # Suppress warning about missing interface methods.

    def __init__(self, oidc_connector: OIDCConnector, scopes: OAuthScopes = OAuthScopes.default) -> None:
        """
        :param oidc_connector: The OIDC connector to use for fetching tokens.
        :param scopes: The OAuth scopes to request.
        """
        self._mutex = asyncio.Lock()

        self._connector = oidc_connector
        self.__token: T | None = None

        assert isinstance(scopes, OAuthScopes), "Scopes must be an instance of OAuthScopes."
        self._scopes: OAuthScopes = scopes

    def _get_token(self) -> T | None:
        """Get the current access token, or None if the token is not available.

        The mutex should be acquired before setting or getting the token.

        :return: The access token, or None if the token is not available.
        """
        assert self._mutex.locked(), "self._mutex should be acquired before getting the token."
        return self.__token

    def _update_token(self, new_token: T) -> None:
        """Update the access token.

        The mutex should be acquired before setting or getting the token.

        :param new_token: The new access token.
        """
        assert self._mutex.locked(), "self._mutex should be acquired before updating the token."
        self.__token = new_token

    @contextlib.asynccontextmanager
    async def _unwrap_token(self) -> AsyncIterator[T]:
        """A context manager that yields the access token if available.

        :raises OAuthError: If the access token is not available.
        """
        async with self._mutex:
            if self._get_token() is None:
                raise OAuthError("Access token is not available.")
            yield self._get_token()

    async def get_default_headers(self) -> HTTPHeaderDict:
        async with self._unwrap_token() as token:
            return HTTPHeaderDict({"Authorization": f"Bearer {token.access_token}"})


class ClientCredentialsAuthorizer(_BaseAuthorizer[AccessToken]):
    """An OAuth authorizer that uses a client credientials to authorize a client.

    The authorizer will automatically refresh the access token when it expires.
    """

    @contextlib.asynccontextmanager
    async def _unwrap_token(self) -> AsyncIterator[T]:
        """A context manager that yields the access token if available, or fetches a new access token if unavailable.

        :raises OAuthError: If the access token cannot be fetched.
        """
        async with self._mutex:
            if self._get_token() is None:
                self._update_token(await self._fetch_token())
            yield self._get_token()

    async def _fetch_token(self) -> AccessToken:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/

        :return: The access token.

        :raises OAuthError: If token cannot be fetched.
        """
        await self._connector.load_config()
        self._validate_configuration()
        data = {  # The payload to send to the OAuth server in the token request.
            "grant_type": "client_credentials",
            "scope": str(self._scopes),
        }

        logger.debug("Fetching access token...")
        return await self._connector.fetch_token(data, AccessToken)

    async def authorize(self) -> None:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/

        :return: The access token.

        :raises OAuthError: If token cannot be fetched.
        """
        async with self._mutex:
            self._update_token(await self._fetch_token())

    async def refresh_token(self) -> bool:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/

        :return: The access token.

        :raises OAuthError: If token cannot be fetched.
        """
        try:
            await self.authorize()
        except Exception:  # noqa
            logger.exception("Failed to refresh the access token.", exc_info=True)
            return False
        else:
            return True

    def _validate_configuration(self):
        """Validate an OIDC configuration against the issuer.

        https://openid.net/specs/openid-connect-discovery-1_0.html#rfc.section.3

        :raises OIDCError: If the configuration is invalid.
        """
        if (
            self._connector.config.grant_types_supported
            and "client_credentials" not in self._connector.config.grant_types_supported
        ):
            raise OIDCError("Authorization provider does not support the 'client_credentials' grant type.")


class AuthorizationCodeAuthorizer(_BaseAuthorizer[UserAccessToken]):
    """An OAuth authorizer that uses a localhost callback to authenticate the user.

    This authorizer launches a local web browser to authenticate the user and obtain an access token. The user is
    required to log in and authorize the application. The authorizer will then use the access token to authorize
    requests.

    The authorizer will automatically refresh the access token when it expires. The refresh token is stored in the
    session and is used to obtain a new access token. If the refresh token expires, the user will need to log in again.
    A mutex is used to ensure that default headers are not returned while the user is logging in or the access token is
    being refreshed.

    This authorizer does not store the access token between sessions. The user will need to log in each time the
    application is started.
    """

    def __init__(
        self, oidc_connector: OIDCConnector, redirect_url: str, scopes: OAuthScopes = OAuthScopes.default
    ) -> None:
        """
        :param oidc_connector: The OIDC connector to use for fetching tokens.
        :param redirect_url: The local URL to redirect the user back to after authorisation.
        :param scopes: The OAuth scopes to request.
        """
        super().__init__(oidc_connector, scopes)
        self._redirect_url = redirect_url

    def _update_token(self, new_token: UserAccessToken) -> None:
        new_token.validate_id_token(issuer=self._connector.issuer, client_id=self._connector.client_id)
        return super()._update_token(new_token)

    async def _handle_login(self, timeout_seconds: int) -> UserAccessToken:
        """Internal method to handle the login process without acquiring the mutex.

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :return: The new access token.

        :raises OAuthError: If the user does not authenticate within the timeout.
        """
        async with OAuthRedirectHandler(self._connector, self._redirect_url) as handler:
            self._validate_configuration()
            return await handler.login(scopes=self._scopes, timeout_seconds=timeout_seconds)

    async def login(self, timeout_seconds: int = 180) -> None:
        """Authenticate the user and obtain an access token.

        This method will launch a web browser to authenticate the user and obtain an access token.

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :raises OAuthError: If the user does not authenticate within the timeout.
        :raises OAuthError: If an error occurred during the authorisation process.
        """
        async with self._mutex:
            new_token = await self._handle_login(timeout_seconds)
            self._update_token(new_token)

    async def refresh_token(self) -> bool:
        async with self._unwrap_token() as old_token, self._connector:
            try:
                if old_token.refresh_token is None:
                    raise OAuthError("Refresh token is missing.")

                data = {  # The payload to send to the OAuth server in the token request.
                    "grant_type": "refresh_token",
                    "refresh_token": old_token.refresh_token,
                }
                logger.debug("Refreshing access token...")
                new_token = await self._connector.fetch_token(data, UserAccessToken)
                self._update_token(new_token)
            except Exception:  # noqa
                logger.exception("Failed to refresh the access token.", exc_info=True)
                return False
            else:
                return True

    def _validate_configuration(self):
        """Validate an OIDC configuration against the issuer.

        https://openid.net/specs/openid-connect-discovery-1_0.html#rfc.section.3

        :raises OIDCError: If the configuration is invalid.
        """
        if "code" not in self._connector.config.response_types_supported:
            raise OIDCError("Authorization provider does not support the 'code' response type.")

        if (
            self._connector.config.grant_types_supported
            and "authorization_code" not in self._connector.config.grant_types_supported
        ):
            raise OIDCError("Authorization provider does not support the 'authorization_code' grant type.")


class DeviceFlowAuthorizer(_BaseAuthorizer[AccessToken]):
    """An OAuth authorizer that uses the device flow to authenticate the user.

    https://datatracker.ietf.org/doc/html/rfc8628

    The device flow is a two-step process. The first step is to begin the device flow and obtain a device code. The
    second step is to wait for the user to authorize the device flow. The access token is then obtained and stored
    in the authorizer.

    After receiving a successful authorization response, the client must display the "user_code" and the
    "verification_uri" to the end user and instruct them to visit the URI in a user agent on another device and
    enter the user code.

    The "verification_uri_complete" value, if available, can be used in any non-textual manner that results in a
    browser being used to open the URI. For example, the URI could be placed in a QR code or NFC tag to save the user
    from manually typing it.

    https://datatracker.ietf.org/doc/html/rfc8628#section-3.3.1
    """

    async def _begin_device_flow(self) -> DeviceFlowResponse:
        """Begin the device flow.

        Send a device authorization request to the authorization server.

        :return: The device authorization response.

        :raises OIDCError: If the authorization provider does not support device flow.
        :raises OAuthError: If the device flow cannot be started.
        """
        await self._connector.load_config()
        self._validate_configuration()
        return await self._connector.begin_device_flow(self._scopes)

    async def _wait_for_authorization(self, flow: DeviceFlowResponse) -> None:
        """Wait for the user to authorize the device flow.

        :param flow: The device authorization response.

        :raises OAuthError: If the user does not authorize the device flow within the required time.
        """
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": flow.device_code,
        }
        try:
            async for handler in flow._retry:
                with handler.suppress_errors():
                    token = await self._connector.fetch_token(data, AccessToken)
        except RetryError:
            raise OAuthError("Failed to authenticate the user.")

        self._update_token(token)

    @contextlib.asynccontextmanager
    async def login(self, timeout_seconds: int = 180) -> AsyncIterator[DeviceFlowResponse]:
        """A context manager to authenticate the user using the device flow.

        The device flow is a two-step process. The first step is to begin the device flow and obtain a device code. The
        second step is to wait for the user to authorize the device flow. The access token is then obtained and stored
        in the authorizer.

        After receiving a successful authorization response, the client must display the "user_code" and the
        "verification_uri" to the end user and instruct them to visit the URI in a user agent on another device and
        enter the user code.

        This authorizer will wait for the user to authorize the device flow upon exiting the context manager.

        https://datatracker.ietf.org/doc/html/rfc8628#section-3.3

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.
            The OAuth provider may specify a shorter timeout in the device authorization response, in which case that
            value will be used.

        :yield: The device authorization response.

        :raises OIDCError: If the authorization provider does not support device flow.
        :raises OAuthError: If the device flow cannot be started.
        :raises OAuthError: If the user does not authorize the device flow within the required time.
        """
        async with self._mutex:
            logger.debug("Starting device flow...")
            flow = await self._begin_device_flow()
            # Override the expires_in value with the timeout_seconds value if it is less than the expires_in value.
            flow.expires_in = min(flow.expires_in, timeout_seconds)
            logger.debug(f"Verification URI: {flow.verification_uri}")
            logger.debug(f"User code: {flow.user_code}")
            yield flow
            await self._wait_for_authorization(flow)

    async def refresh_token(self) -> bool:
        logger.exception("Unable to refresh the access token because refresh tokens are not supported for device flow.")
        return False

    def _validate_configuration(self):
        if (
            self._connector.config.grant_types_supported
            and "urn:ietf:params:oauth:grant-type:device_code" not in self._connector.config.grant_types_supported
        ):
            raise OIDCError("Authorization provider does not support the 'device_code' grant type.")
